@@ -1,36 +1,39 @@
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
- *
- * This file is part of libxls -- A multiplatform, C/C++ library
- * for parsing Excel(TM) files.
- *
- * Redistribution and use in source and binary forms, with or without modification, are
- * permitted provided that the following conditions are met:
- *
- *    1. Redistributions of source code must retain the above copyright notice, this list of
- *       conditions and the following disclaimer.
- *
- *    2. Redistributions in binary form must reproduce the above copyright notice, this list
- *       of conditions and the following disclaimer in the documentation and/or other materials
- *       provided with the distribution.
- *
- * THIS SOFTWARE IS PROVIDED BY David Hoerl ''AS IS'' AND ANY EXPRESS OR IMPLIED
- * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
- * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL David Hoerl OR
- * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
- * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
- * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
- * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
- * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
- * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
  * Copyright 2004 Komarov Valery
  * Copyright 2006 Christophe Leitienne
+ * Copyright 2008-2017 David Hoerl
  * Copyright 2013 Bob Colbert
- * Copyright 2008-2013 David Hoerl
+ * Copyright 2013-2018 Evan Miller
+ *
+ * This file is part of libxls -- A multiplatform, C/C++ library for parsing
+ * Excel(TM) files.
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions are met:
+ *
+ *    1. Redistributions of source code must retain the above copyright notice,
+ *    this list of conditions and the following disclaimer.
+ *
+ *    2. Redistributions in binary form must reproduce the above copyright
+ *    notice, this list of conditions and the following disclaimer in the
+ *    documentation and/or other materials provided with the distribution.
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ''AS
+ * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
+ * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
+ * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR
+ * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
+ * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
+ * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
+ * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
+ * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
+ * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
+ * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
  */
 
-#include "config.h"
+#include "config.h" 
 
 #include <memory.h>
 #include <math.h>
@@ -53,8 +56,8 @@ extern int xls_debug;
 static const DWORD ENDOFCHAIN	= 0xFFFFFFFE;	// -2
 static const DWORD FREESECT		= 0xFFFFFFFF;	// -1
 
-static size_t sector_pos(OLE2* ole2, size_t sid);
-static ssize_t sector_read(OLE2* ole2, void *buffer, size_t sid);
+static size_t sector_pos(OLE2* ole2, DWORD sid);
+static ssize_t sector_read(OLE2* ole2, void *buffer, DWORD sid);
 static ssize_t read_MSAT(OLE2* ole2, OLE2Header *oleh);
 static void *ole_malloc(size_t len);
 
@@ -65,8 +68,24 @@ static void *ole_malloc(size_t len) {
     return malloc(len);
 }
 
+int ole2_validate_sector(DWORD sector, OLE2 *ole) {
+    if (sector >= ole->SecIDCount) {
+        if (xls_debug) fprintf(stderr, "Error: fatpos %d out-of-bounds for SecID[%d]\n",
+                (int)sector, ole->SecIDCount);
+        return 0;
+    }
+
+    if (sector == xlsIntVal(ole->SecID[sector])) {
+        if (xls_debug) fprintf(stderr, "Error: Sector loop detected, SecID[%d] = %d\n",
+                (int)sector, (int)sector);
+        return 0;
+    }
+
+    return 1;
+}
+
 // Read next sector of stream
-int ole2_bufread(OLE2Stream* olest)
+int ole2_bufread(OLE2Stream* olest) 
 {
 	BYTE *ptr;
 
@@ -80,15 +99,15 @@ int ole2_bufread(OLE2Stream* olest)
                 return -1;
 
             if (olest->fatpos*olest->ole->lssector + olest->bufsize > olest->ole->SSATCount) {
-                if (xls_debug) fprintf3(stderr, "Error: fatpos %d out-of-bounds for SSAT\n", (int)olest->fatpos);
+                if (xls_debug) fprintf(stderr, "Error: fatpos %d out-of-bounds for SSAT\n", (int)olest->fatpos);
                 return -1;
             }
 
 			ptr = olest->ole->SSAT + olest->fatpos*olest->ole->lssector;
-			memcpy(olest->buf, ptr, olest->bufsize);
+			memcpy(olest->buf, ptr, olest->bufsize); 
 
             if (olest->fatpos >= olest->ole->SSecIDCount) {
-                if (xls_debug) fprintf4(stderr, "Error: fatpos %d out-of-bounds for SSecID[%d]\n",
+                if (xls_debug) fprintf(stderr, "Error: fatpos %d out-of-bounds for SSecID[%d]\n",
                         (int)olest->fatpos, olest->ole->SSecIDCount);
                 return -1;
             }
@@ -99,19 +118,11 @@ int ole2_bufread(OLE2Stream* olest)
 		} else {
 			if ((int)olest->fatpos < 0 ||
                 sector_read(olest->ole, olest->buf, olest->fatpos) == -1) {
-                if (xls_debug) fprintf3(stderr, "Error: Unable to read sector #%d\n", (int)olest->fatpos);
+                if (xls_debug) fprintf(stderr, "Error: Unable to read sector #%d\n", (int)olest->fatpos);
                 return -1;
             }
 
-            if (olest->fatpos >= olest->ole->SecIDCount) {
-                if (xls_debug) fprintf4(stderr, "Error: fatpos %d out-of-bounds for SecID[%d]\n",
-                        (int)olest->fatpos, olest->ole->SecIDCount);
-                return -1;
-            }
-
-            if (olest->fatpos == xlsIntVal(olest->ole->SecID[olest->fatpos])) {
-                if (xls_debug) fprintf4(stderr, "Error: Sector loop detected, SecID[%d] = %d\n",
-                        (int)olest->fatpos, (int)olest->fatpos);
+            if (!ole2_validate_sector(olest->fatpos, olest->ole)) {
                 return -1;
             }
 
@@ -138,7 +149,7 @@ ssize_t ole2_read(void* buf, size_t size, size_t count, OLE2Stream* olest)
     if ((long)olest->size>=0 && !olest->sfat)	// directory is -1
     {
 		size_t rem;
-		rem = olest->size - (olest->cfat*olest->ole->lsector+olest->pos);
+		rem = olest->size - (olest->cfat*olest->ole->lsector+olest->pos);		
         totalReadCount = rem<totalReadCount?rem:totalReadCount;
         if (rem<=0) olest->eof=1;
 
@@ -176,15 +187,15 @@ ssize_t ole2_read(void* buf, size_t size, size_t count, OLE2Stream* olest)
 	// printf("=====\n");
 
 #ifdef OLE_DEBUG
-    fprintf2(stderr, "----------------------------------------------\n");
-    fprintf2(stderr, "ole2_read (end)\n");
-    fprintf3(stderr, "start:		%d \n",olest->start);
-    fprintf3(stderr, "pos:		%d \n",(int)olest->pos);
-    fprintf3(stderr, "cfat:		%d \n",(int)olest->cfat);
-    fprintf3(stderr, "size:		%d \n",(int)olest->size);
-    fprintf3(stderr, "fatpos:		%d \n",(int)olest->fatpos);
-    fprintf3(stderr, "bufsize:		%d \n",(int)olest->bufsize);
-    fprintf3(stderr, "eof:		%d \n",olest->eof);
+    fprintf(stderr, "----------------------------------------------\n");
+    fprintf(stderr, "ole2_read (end)\n");
+    fprintf(stderr, "start:		%d \n",olest->start);
+    fprintf(stderr, "pos:		%d \n",(int)olest->pos);
+    fprintf(stderr, "cfat:		%d \n",(int)olest->cfat);
+    fprintf(stderr, "size:		%d \n",(int)olest->size);
+    fprintf(stderr, "fatpos:		%d \n",(int)olest->fatpos);
+    fprintf(stderr, "bufsize:		%d \n",(int)olest->bufsize);
+    fprintf(stderr, "eof:		%d \n",olest->eof);
 #endif
 
     return didReadCount;
@@ -196,8 +207,8 @@ OLE2Stream* ole2_sopen(OLE2* ole,DWORD start, size_t size)
     OLE2Stream* olest=NULL;
 
 #ifdef OLE_DEBUG
-    fprintf2(stderr, "----------------------------------------------\n");
-    fprintf3(stderr, "ole2_sopen start=%Xh\n", start);
+    fprintf(stderr, "----------------------------------------------\n");
+    fprintf(stderr, "ole2_sopen start=%Xh\n", start);
 #endif
 
 	olest = calloc(1, sizeof(OLE2Stream));
@@ -227,7 +238,7 @@ OLE2Stream* ole2_sopen(OLE2* ole,DWORD start, size_t size)
 int ole2_seek(OLE2Stream* olest,DWORD ofs)
 {
 #ifdef OLE_DEBUG
-    fprintf3(stderr, "SEEK %x\n", ofs);
+    fprintf(stderr, "SEEK %x\n", ofs);
 #endif
 	if(olest->sfat) {
 		ldiv_t div_rez=ldiv(ofs,olest->ole->lssector);
@@ -254,14 +265,14 @@ int ole2_seek(OLE2Stream* olest,DWORD ofs)
 		ldiv_t div_rez=ldiv(ofs,olest->ole->lsector);
 		int i;
 #ifdef OLE_DEBUG
-        fprintf4(stderr, "seeking fatpos%lu start %u\n", olest->fatpos, olest->start);
+        fprintf(stderr, "seeking fatpos%lu start %u\n", olest->fatpos, olest->start);
 #endif
 		olest->fatpos=olest->start;
 
         if (div_rez.quot!=0)
         {
             for (i=0;i<div_rez.quot;i++) {
-                if (olest->fatpos >= olest->ole->SecIDCount)
+                if (!ole2_validate_sector(olest->fatpos, olest->ole))
                     return -1;
                 olest->fatpos=xlsIntVal(olest->ole->SecID[olest->fatpos]);
             }
@@ -284,15 +295,15 @@ OLE2Stream*  ole2_fopen(OLE2* ole, const char *file)
     int i;
 
 #ifdef OLE_DEBUG
-    fprintf2(stderr, "----------------------------------------------\n");
-    fprintf3(stderr, "ole2_fopen %s\n", file);
+    fprintf(stderr, "----------------------------------------------\n");
+    fprintf(stderr, "ole2_fopen %s\n", file);
 #endif
 
     for (i=0;i<ole->files.count;i++) {
 		char *str = ole->files.file[i].name;
 #ifdef OLE_DEBUG
-		fprintf2(stderr, "----------------------------------------------\n");
-		fprintf3(stderr, "ole2_fopen found %s\n", str);
+		fprintf(stderr, "----------------------------------------------\n");
+		fprintf(stderr, "ole2_fopen found %s\n", str);
 #endif
         if (str && strcmp(str,file)==0)	// newer versions of Excel don't write the "Root Entry" string for the first set of data
         {
@@ -342,7 +353,7 @@ static ssize_t ole2_read_header(OLE2 *ole) {
 
 	// make sure the file looks good. Note: this code only works on Little Endian machines
 	if(oleh->id[0] != 0xE011CFD0 || oleh->id[1] != 0xE11AB1A1 || oleh->byteorder != 0xFFFE) {
-        if (xls_debug) fprintf2(stderr, "Not an excel file\n");
+        if (xls_debug) fprintf(stderr, "Not an excel file\n");
         total_bytes_read = -1;
         goto cleanup;
 	}
@@ -353,11 +364,11 @@ static ssize_t ole2_read_header(OLE2 *ole) {
     ole->lssector=64;
 
 	if (oleh->lsectorB != 9 || oleh->lssectorB != 6) {	// 2**9 == 512, 2**6 == 64
-        if (xls_debug) fprintf2(stderr, "Unexpected sector size\n");
+        if (xls_debug) fprintf(stderr, "Unexpected sector size\n");
         total_bytes_read = -1;
         goto cleanup;
     }
-
+	
     ole->cfat=oleh->cfat;
     ole->dirstart=oleh->dirstart;
     ole->sectorcutoff=oleh->sectorcutoff;
@@ -368,23 +379,23 @@ static ssize_t ole2_read_header(OLE2 *ole) {
     ole->files.count=0;
 
 #ifdef OLE_DEBUG
-		fprintf2(stderr, "==== OLE HEADER ====\n");
+		fprintf(stderr, "==== OLE HEADER ====\n");
 		//printf ("Header Size:   %i \n", sizeof(OLE2Header));
 		//printf ("id[0]-id[1]:   %X-%X \n", oleh->id[0], oleh->id[1]);
-		fprintf3(stderr, "verminor:      %X \n",oleh->verminor);
-		fprintf3(stderr, "verdll:        %X \n",oleh->verdll);
+		fprintf(stderr, "verminor:      %X \n",oleh->verminor);
+		fprintf(stderr, "verdll:        %X \n",oleh->verdll);
 		//printf ("Byte order:    %X \n",oleh->byteorder);
-		fprintf4(stderr, "sect len:      %X (%i)\n",ole->lsector,ole->lsector);		// ole
-		fprintf4(stderr, "mini len:      %X (%i)\n",ole->lssector,ole->lssector);	// ole
-		fprintf3(stderr, "Fat sect.:     %i \n",oleh->cfat);
-		fprintf3(stderr, "Dir Start:     %i \n",oleh->dirstart);
-
-		fprintf3(stderr, "Mini Cutoff:   %i \n",oleh->sectorcutoff);
-		fprintf3(stderr, "MiniFat Start: %X \n",oleh->sfatstart);
-		fprintf3(stderr, "Count MFat:    %i \n",oleh->csfat);
-		fprintf3(stderr, "Dif start:     %X \n",oleh->difstart);
-		fprintf3(stderr, "Count Dif:     %i \n",oleh->cdif);
-		fprintf4(stderr, "Fat Size:      %u (0x%X) \n",oleh->cfat*ole->lsector,oleh->cfat*ole->lsector);
+		fprintf(stderr, "sect len:      %X (%i)\n",ole->lsector,ole->lsector);		// ole
+		fprintf(stderr, "mini len:      %X (%i)\n",ole->lssector,ole->lssector);	// ole
+		fprintf(stderr, "Fat sect.:     %i \n",oleh->cfat);
+		fprintf(stderr, "Dir Start:     %i \n",oleh->dirstart);
+		
+		fprintf(stderr, "Mini Cutoff:   %i \n",oleh->sectorcutoff);
+		fprintf(stderr, "MiniFat Start: %X \n",oleh->sfatstart);
+		fprintf(stderr, "Count MFat:    %i \n",oleh->csfat);
+		fprintf(stderr, "Dif start:     %X \n",oleh->difstart);
+		fprintf(stderr, "Count Dif:     %i \n",oleh->cdif);
+		fprintf(stderr, "Fat Size:      %u (0x%X) \n",oleh->cfat*ole->lsector,oleh->cfat*ole->lsector);
 #endif
     // read directory entries
     if ((bytes_read = read_MSAT(ole, oleh)) == -1) {
@@ -423,44 +434,44 @@ static ssize_t ole2_read_body(OLE2 *ole) {
             goto cleanup;
         }
         name=unicode_decode(pss->name, pss->bsize, 0, "UTF-8");
-#ifdef OLE_DEBUG
-		fprintf4(stderr, "OLE NAME: %s count=%d\n", name, (int)ole->files.count);
+#ifdef OLE_DEBUG	
+		fprintf(stderr, "OLE NAME: %s count=%d\n", name, (int)ole->files.count);
 #endif
-        if (pss->type == PS_USER_ROOT || pss->type == PS_USER_STREAM) // (name!=NULL) //
+        if (pss->type == PS_USER_ROOT || pss->type == PS_USER_STREAM) // (name!=NULL) // 
         {
 
-#ifdef OLE_DEBUG
-			fprintf3(stderr, "OLE TYPE: %s file=%d \n", pss->type == PS_USER_ROOT ? "root" : "user", (int)ole->files.count);
-#endif
+#ifdef OLE_DEBUG		
+			fprintf(stderr, "OLE TYPE: %s file=%d \n", pss->type == PS_USER_ROOT ? "root" : "user", (int)ole->files.count);
+#endif		
             ole->files.file = realloc(ole->files.file,(ole->files.count+1)*sizeof(struct st_olefiles_data));
             ole->files.file[ole->files.count].name=name;
             ole->files.file[ole->files.count].start=pss->sstart;
             ole->files.file[ole->files.count].size=pss->size;
             ole->files.count++;
-
+			
 			if(pss->sstart == ENDOFCHAIN) {
 				if (xls_debug) verbose("END OF CHAIN\n");
 			} else if(pss->type == PS_USER_STREAM) {
 #ifdef OLE_DEBUG
-					fprintf2(stderr, "----------------------------------------------\n");
-					fprintf5(stderr, "name: %s (size=%d [c=%c])\n", name, pss->bsize, name ? name[0]:' ');
-					fprintf3(stderr, "bsize %i\n",pss->bsize);
-					fprintf3(stderr, "type %i\n",pss->type);
-					fprintf3(stderr, "flag %i\n",pss->flag);
-					fprintf3(stderr, "left %X\n",pss->left);
-					fprintf3(stderr, "right %X\n",pss->right);
-					fprintf3(stderr, "child %X\n",pss->child);
-					fprintf10(stderr, "guid %.4X-%.4X-%.4X-%.4X %.4X-%.4X-%.4X-%.4X\n",
+					fprintf(stderr, "----------------------------------------------\n");
+					fprintf(stderr, "name: %s (size=%d [c=%c])\n", name, pss->bsize, name ? name[0]:' ');
+					fprintf(stderr, "bsize %i\n",pss->bsize);
+					fprintf(stderr, "type %i\n",pss->type);
+					fprintf(stderr, "flag %i\n",pss->flag);
+					fprintf(stderr, "left %X\n",pss->left);
+					fprintf(stderr, "right %X\n",pss->right);
+					fprintf(stderr, "child %X\n",pss->child);
+					fprintf(stderr, "guid %.4X-%.4X-%.4X-%.4X %.4X-%.4X-%.4X-%.4X\n",
                             pss->guid[0],pss->guid[1],pss->guid[2],pss->guid[3],
 						pss->guid[4],pss->guid[5],pss->guid[6],pss->guid[7]);
-					fprintf3(stderr, "user flag %.4X\n",pss->userflags);
-					fprintf3(stderr, "sstart %.4d\n",pss->sstart);
-					fprintf3(stderr, "size %.4d\n",pss->size);
+					fprintf(stderr, "user flag %.4X\n",pss->userflags);
+					fprintf(stderr, "sstart %.4d\n",pss->sstart);
+					fprintf(stderr, "size %.4d\n",pss->size);
 #endif
 			} else if(pss->type == PS_USER_ROOT) {
 				DWORD sector, k, blocks;
 				BYTE *wptr;
-
+				
 				blocks = (pss->size + (ole->lsector - 1)) / ole->lsector;	// count partial
 				if ((ole->SSAT = ole_malloc(blocks*ole->lsector)) == NULL) {
                     total_bytes_read = -1;
@@ -474,7 +485,11 @@ static ssize_t ole2_read_body(OLE2 *ole) {
 				for(k=0; k<blocks; ++k) {
 					// printf("block %d sector %d\n", k, sector);
                     if (sector == ENDOFCHAIN || sector_read(ole, wptr, sector) == -1) {
-                        if (xls_debug) fprintf3(stderr, "Unable to read sector #%d\n", sector);
+                        if (xls_debug) fprintf(stderr, "Unable to read sector #%d\n", sector);
+                        total_bytes_read = -1;
+                        goto cleanup;
+                    }
+                    if (!ole2_validate_sector(sector, ole)) {
                         total_bytes_read = -1;
                         goto cleanup;
                     }
@@ -482,7 +497,7 @@ static ssize_t ole2_read_body(OLE2 *ole) {
 					wptr += ole->lsector;
 					sector = xlsIntVal(ole->SecID[sector]);
 				}
-			}
+			}	
 		} else {
 			free(name);
 		}
@@ -523,15 +538,15 @@ OLE2* ole2_open_file(const char *file)
     OLE2* ole = NULL;
 
 #ifdef OLE_DEBUG
-    fprintf2(stderr, "----------------------------------------------\n");
-    fprintf3(stderr, "ole2_open_file %s\n", file);
+    fprintf(stderr, "----------------------------------------------\n");
+    fprintf(stderr, "ole2_open_file %s\n", file);
 #endif
 
 	if(xls_debug) printf("ole2_open: %s\n", file);
     ole=(OLE2*)calloc(1, sizeof(OLE2));
 
     if (!(ole->file=fopen(file, "rb"))) {
-        if(xls_debug) fprintf2(stderr, "File not found\n");
+        if(xls_debug) fprintf(stderr, "File not found\n");
         free(ole);
         return NULL;
     }
@@ -555,14 +570,14 @@ void ole2_close(OLE2* ole2)
     if (ole2->file)
         fclose(ole2->file);
 
-	for(i=0; i<ole2->files.count; ++i) {
-		free(ole2->files.file[i].name);
-	}
-	free(ole2->files.file);
-	free(ole2->SecID);
-	free(ole2->SSecID);
-	free(ole2->SSAT);
-	free(ole2);
+    for(i=0; i<ole2->files.count; ++i) {
+        free(ole2->files.file[i].name);
+    }
+    free(ole2->files.file);
+    free(ole2->SecID);
+    free(ole2->SSecID);
+    free(ole2->SSAT);
+    free(ole2);
 }
 
 void ole2_fclose(OLE2Stream* ole2st)
@@ -572,23 +587,25 @@ void ole2_fclose(OLE2Stream* ole2st)
 }
 
 // Return offset in bytes of a sector from its sid
-static size_t sector_pos(OLE2* ole2, size_t sid)
+static size_t sector_pos(OLE2* ole2, DWORD sid)
 {
     return 512 + sid * ole2->lsector;
 }
 // Read one sector from its sid
-static ssize_t sector_read(OLE2* ole2, void *buffer, size_t sid)
+static ssize_t sector_read(OLE2* ole2, void *buffer, DWORD sid)
 {
 	size_t num;
 	size_t seeked;
 
 	if ((seeked = ole2_fseek(ole2, sector_pos(ole2, sid))) != 0) {
-		if (xls_debug) fprintf5(stderr, "Error: wanted to seek to sector %zu (0x%zx) loc=%zu\n", sid, sid, sector_pos(ole2, sid));
+		if (xls_debug) fprintf(stderr, "Error: wanted to seek to sector %u (0x%x) loc=%u\n", sid, sid,
+                (unsigned int)sector_pos(ole2, sid));
         return -1;
     }
 
     if ((num = ole2_fread(ole2, buffer, ole2->lsector, 1)) != 1) {
-        if (xls_debug) fprintf4(stderr, "Error: fread wanted 1 got %zu loc=%zu\n", num, sector_pos(ole2, sid));
+        if (xls_debug) fprintf(stderr, "Error: fread wanted 1 got %lu loc=%u\n", (unsigned long)num,
+                (unsigned int)sector_pos(ole2, sid));
         return -1;
     }
 
@@ -596,15 +613,15 @@ static ssize_t sector_read(OLE2* ole2, void *buffer, size_t sid)
 }
 
 // read first 109 sectors of MSAT from header
-static ssize_t read_MSAT_header(OLE2* ole2, OLE2Header* oleh, int sectorCount) {
+static ssize_t read_MSAT_header(OLE2* ole2, OLE2Header* oleh, DWORD sectorCount) {
     BYTE *sector = (BYTE*)ole2->SecID;
     ssize_t bytes_read = 0, total_bytes_read = 0;
-    int sectorNum;
+    DWORD sectorNum;
 
-    for (sectorNum = 0; sectorNum < sectorCount; sectorNum++)
+    for (sectorNum = 0; sectorNum < sectorCount && sectorNum < 109; sectorNum++)
     {
         if ((bytes_read = sector_read(ole2, sector, oleh->MSAT[sectorNum])) == -1) {
-            if (xls_debug) fprintf3(stderr, "Error: Unable to read sector #%d\n", oleh->MSAT[sectorNum]);
+            if (xls_debug) fprintf(stderr, "Error: Unable to read sector #%d\n", oleh->MSAT[sectorNum]);
             return -1;
         }
         sector += ole2->lsector;
@@ -614,12 +631,12 @@ static ssize_t read_MSAT_header(OLE2* ole2, OLE2Header* oleh, int sectorCount) {
 }
 
 // Add additional sectors of the MSAT
-static ssize_t read_MSAT_body(OLE2 *ole2, int sectorOffset, int sectorCount) {
+static ssize_t read_MSAT_body(OLE2 *ole2, DWORD sectorOffset, DWORD sectorCount) {
     DWORD sid = ole2->difstart;
     ssize_t bytes_read = 0, total_bytes_read = 0;
-    int sectorNum = sectorOffset;
+    DWORD sectorNum = sectorOffset;
 
-    BYTE *sector = ole_malloc(ole2->lsector);
+    DWORD *sector = ole_malloc(ole2->lsector);
     //printf("sid=%u (0x%x) sector=%u\n", sid, sid, ole2->lsector);
     while (sid != ENDOFCHAIN && sid != FREESECT) // FREESECT only here due to an actual file that requires it (old Apple Numbers bug)
     {
@@ -627,7 +644,7 @@ static ssize_t read_MSAT_body(OLE2 *ole2, int sectorOffset, int sectorCount) {
         // read MSAT sector
         if ((bytes_read = sector_read(ole2, sector, sid)) == -1) {
             total_bytes_read = -1;
-            if (xls_debug) fprintf3(stderr, "Error: Unable to read sector #%d\n", sid);
+            if (xls_debug) fprintf(stderr, "Error: Unable to read sector #%d\n", sid);
             goto cleanup;
         }
         total_bytes_read += bytes_read;
@@ -635,18 +652,18 @@ static ssize_t read_MSAT_body(OLE2 *ole2, int sectorOffset, int sectorCount) {
         // read content
         for (posInSector = 0; posInSector < (ole2->lsector-4)/4; posInSector++)
         {
-            DWORD s = *(DWORD_UA *)(sector + posInSector*4);
+            DWORD s = sector[posInSector];
             //printf("   s[%d]=%d (0x%x)\n", posInSector, s, s);
 
             if (s != ENDOFCHAIN && s != FREESECT) // see patch in Bug 31. For very large files
             {
                 if (sectorNum == sectorCount) {
-                    if (xls_debug) fprintf3(stderr, "Error: Unable to seek to sector #%d\n", s);
+                    if (xls_debug) fprintf(stderr, "Error: Unable to seek to sector #%d\n", s);
                     total_bytes_read = -1;
                     goto cleanup;
                 }
                 if ((bytes_read = sector_read(ole2, (BYTE*)(ole2->SecID)+sectorNum*ole2->lsector, s)) == -1) {
-                    if (xls_debug) fprintf3(stderr, "Error: Unable to read sector #%d\n", s);
+                    if (xls_debug) fprintf(stderr, "Error: Unable to read sector #%d\n", s);
                     total_bytes_read = -1;
                     goto cleanup;
                 }
@@ -654,7 +671,7 @@ static ssize_t read_MSAT_body(OLE2 *ole2, int sectorOffset, int sectorCount) {
                 sectorNum++;
             }
         }
-        sid = *(DWORD_UA *)(sector + posInSector*4);
+        sid = sector[posInSector];
         //printf("   s[%d]=%d (0x%x)\n", posInSector, sid, sid);
     }
 #ifdef OLE_DEBUG
@@ -690,6 +707,10 @@ static ssize_t read_MSAT_trailer(OLE2 *ole2) {
 			if (sector == ENDOFCHAIN || sector_read(ole2, wptr, sector) == -1) {
                 total_bytes_read = -1;
                 goto cleanup;
+            }
+            if (!ole2_validate_sector(sector, ole2)) {
+                total_bytes_read = -1;
+                goto cleanup;
             }
 			wptr += ole2->lsector;
             total_bytes_read += ole2->lsector;
@@ -699,7 +720,7 @@ static ssize_t read_MSAT_trailer(OLE2 *ole2) {
 		if(xls_debug) {
 			int i;
 			for(i=0; i<ole2->csfat; ++i) {
-				if(ole2->SSecID[i] != FREESECT) fprintf4(stderr, "SSecID[%d]=%d\n", i, ole2->SSecID[i]);
+				if(ole2->SSecID[i] != FREESECT) fprintf(stderr, "SSecID[%d]=%d\n", i, ole2->SSecID[i]);
 			}
 		}
 #endif
@@ -714,17 +735,20 @@ cleanup:
 static ssize_t read_MSAT(OLE2* ole2, OLE2Header* oleh)
 {
     // reconstitution of the MSAT
-    int count = (ole2->cfat < 109) ? ole2->cfat : 109;
-    if(count <= 0) {
-        if (xls_debug) fprintf2(stderr, "Error: MSAT count out-of-bounds\n");
+    DWORD count = ole2->cfat;
+    if(count == 0 || count > (1 << 24)) {
+        if (xls_debug) fprintf(stderr, "Error: MSAT count %u out-of-bounds\n", count);
         return -1;
     }
 
     ssize_t total_bytes_read = 0;
     ssize_t bytes_read = 0;
 
-    ole2->SecID = ole_malloc(count*ole2->lsector);
     ole2->SecIDCount = count*ole2->lsector/4;
+    if ((ole2->SecID = ole_malloc(ole2->SecIDCount * sizeof(DWORD))) == NULL) {
+        total_bytes_read = -1;
+        goto cleanup;
+    }
 
     if ((bytes_read = read_MSAT_header(ole2, oleh, count)) == -1) {
         total_bytes_read = -1;
